@@ -1,6 +1,7 @@
-#' @importFrom future Future nbrOfWorkers future resolve value as.FutureGlobals getGlobalsAndPackages
+#' @importFrom future cancel Future nbrOfWorkers future resolve value as.FutureGlobals getGlobalsAndPackages FutureInterruptError
 future_xapply <- local({
   tmpl_expr_options <- bquote_compile({
+    "# future.apply:::future_xapply(): preserve future option"
     ...future.globals.maxSize.org <- getOption("future.globals.maxSize")
     if (!identical(...future.globals.maxSize.org, ...future.globals.maxSize)) {
       oopts <- options(future.globals.maxSize = ...future.globals.maxSize)
@@ -8,8 +9,13 @@ future_xapply <- local({
     }
     .(expr)
   })
-  
-  function(FUN, nX, chunk_args, args = NULL, MoreArgs = NULL, expr, envir = parent.frame(), future.envir, future.globals, future.packages, future.scheduling, future.chunk.size, future.stdout, future.conditions, future.seed, future.label, get_chunk, fcn_name, args_name, ..., debug) {
+
+  function(FUN, nX, chunk_args, args = NULL, MoreArgs = NULL, expr, envir = parent.frame(), future.envir, future.globals, future.packages, future.scheduling, future.chunk.size, future.stdout, future.conditions, future.seed, future.label, get_chunk, fcn_name = "future_xapply", args_name, ..., debug) {
+    if (debug) {
+      mdebugf_push("%s() -> future_xapply() ...", fcn_name)
+      on.exit(mdebug_pop())
+    }
+      
     stop_if_not(is.function(FUN))
     
     stop_if_not(is.logical(future.stdout), length(future.stdout) == 1L)
@@ -96,6 +102,10 @@ future_xapply <- local({
   
     ## At this point a globals should be resolved and we should know their total size
   ##  stop_if_not(attr(globals, "resolved"), !is.na(attr(globals, "total_size")))
+    if (debug) {
+      mdebugf("Globals pass to each chunk: [n=%d] %s", length(globals), commaq(names(globals)))
+      mstr(globals)
+    }
   
     ## To please R CMD check
     ...future.FUN <- ...future.elements_ii <- ...future.seeds_ii <-
@@ -104,7 +114,7 @@ future_xapply <- local({
     globals.maxSize <- getOption("future.globals.maxSize")
     globals.maxSize.default <- globals.maxSize
     if (is.null(globals.maxSize.default)) globals.maxSize.default <- 500 * 1024^2
-    
+
     nchunks <- length(chunks)
     if (debug) mdebugf("Number of futures (= number of chunks): %d", nchunks)
 
@@ -137,148 +147,165 @@ future_xapply <- local({
   
     if (debug) mdebugf("Launching %d futures (chunks) ...", nchunks)
     fs <- vector("list", length = nchunks)
-    for (ii in seq_along(chunks)) {
-      chunk <- chunks[[ii]]
-      if (debug) mdebugf("Chunk #%d of %d ...", ii, length(chunks))
+    values <- tryCatch({
+      for (ii in seq_along(chunks)) {
+        chunk <- chunks[[ii]]
+        if (debug) mdebugf("Chunk #%d of %d ...", ii, length(chunks))
+    
+        args_ii <- get_chunk(chunk_args, chunk)
+        globals_ii <- globals
+        ## Subsetting outside future is more efficient
+        globals_ii[["...future.elements_ii"]] <- args_ii
+        packages_ii <- packages
+    
+        if (scanForGlobals) {
+          if (debug) mdebugf(" - Finding globals in '%s' for chunk #%d ...", args_name, ii)
+          gp <- getGlobalsAndPackages(args_ii, envir = envir, globals = TRUE)
+          globals_args <- gp$globals
+          packages_args <- gp$packages
+          gp <- NULL
+          
+          if (debug) {
+            mdebugf("   + additional globals found: [n=%d] %s",
+                    length(globals_args), commaq(names(globals_args)))
+            mdebugf("   + additional namespaces needed: [n=%d] %s",
+                    length(packages_args), commaq(packages_args))
+          }
+        
+          ## Export also globals found in arguments?
+          if (length(globals_args) > 0L) {
+            reserved <- intersect(c("...future.FUN", "...future.elements_ii",
+                                    "...future.seeds_ii"), names(globals_args))
+            if (length(reserved) > 0) {
+              stop("Detected globals in '%s' using reserved variables names: ",
+                   args_name, commaq(reserved))
+            }
+            globals_args <- as.FutureGlobals(globals_args)
+            globals_ii <- unique(c(globals_ii, globals_args))
+    
+            ## Packages needed due to globals in arguments?
+            if (length(packages_args) > 0L)
+              packages_ii <- unique(c(packages_ii, packages_args))
+          }
+          
+          if (debug) mdebugf(" - Finding globals in '%s' for chunk #%d ... DONE", args_name, ii)
+        }
+        
+        args_ii <- NULL    
+    ##    stop_if_not(attr(globals_ii, "resolved"))
   
-      args_ii <- get_chunk(chunk_args, chunk)
-      globals_ii <- globals
-      ## Subsetting outside future is more efficient
-      globals_ii[["...future.elements_ii"]] <- args_ii
-      packages_ii <- packages
+        if (!is.null(globals.maxSize)) {
+          globals_ii["...future.globals.maxSize"] <- list(globals.maxSize)
+        }
   
-      if (scanForGlobals) {
-        if (debug) mdebugf(" - Finding globals in '%s' for chunk #%d ...", args_name, ii)
-        gp <- getGlobalsAndPackages(args_ii, envir = envir, globals = TRUE)
-        globals_args <- gp$globals
-        packages_args <- gp$packages
-        gp <- NULL
+        ## Adjust option 'future.globals.maxSize' to account for the fact that more
+        ## than one element is processed per future.  The adjustment is done by
+        ## scaling up the limit by the number of elements in the chunk.  This is
+        ## a "good enough" approach.
+        ## (https://github.com/futureverse/future.apply/issues/8).
+        if (length(chunks) > 1L) {
+          options(future.globals.maxSize = length(chunks) * globals.maxSize.default)
+          if (debug) mdebugf(" - Adjusted option 'future.globals.maxSize': %.0f -> %d * %.0f = %.0f (bytes)", globals.maxSize.default, length(chunks), globals.maxSize.default, getOption("future.globals.maxSize"))
+          on.exit(options(future.globals.maxSize = globals.maxSize), add = TRUE)
+        }
+        
+        ## Using RNG seeds or not?
+        if (is.null(seeds)) {
+          if (debug) mdebug(" - seeds: <none>")
+        } else {
+          if (debug) mdebugf(" - seeds: [%d] <seeds>", length(chunk))
+          globals_ii[["...future.seeds_ii"]] <- seeds[chunk]
+        }
+  
+        if (debug) {
+          mdebugf(" - All globals exported: [n=%d] %s",
+                  length(globals_ii), commaq(names(globals_ii)))
+        }
+
+        ## FIXME: Handle interrupts also here, i.e. as soon as we have
+        ## launched the first future, we should be able to interrupt it
+        fs[[ii]] <- future(
+          expr, substitute = FALSE,
+          envir = future.envir,
+          stdout = future.stdout,
+          conditions = future.conditions,
+          globals = globals_ii, packages = packages_ii,
+          seed = future.seed,
+          label = labels[ii]
+        )
         
         if (debug) {
-          mdebugf("   + additional globals found: [n=%d] %s",
-                  length(globals_args), commaq(names(globals_args)))
-          mdebugf("   + additional namespaces needed: [n=%d] %s",
-                  length(packages_args), commaq(packages_args))
-        }
-      
-        ## Export also globals found in arguments?
-        if (length(globals_args) > 0L) {
-          reserved <- intersect(c("...future.FUN", "...future.elements_ii",
-                                  "...future.seeds_ii"), names(globals_args))
-          if (length(reserved) > 0) {
-            stop("Detected globals in '%s' using reserved variables names: ",
-                 args_name, paste(sQuote(reserved), collapse = ", "))
-          }
-          globals_args <- as.FutureGlobals(globals_args)
-          globals_ii <- unique(c(globals_ii, globals_args))
-  
-          ## Packages needed due to globals in arguments?
-          if (length(packages_args) > 0L)
-            packages_ii <- unique(c(packages_ii, packages_args))
+          mdebug("Created future:")
+          mprint(fs[[ii]])
         }
         
-        if (debug) mdebugf(" - Finding globals in '%s' for chunk #%d ... DONE", args_name, ii)
-      }
-      
-      args_ii <- NULL    
-  ##    stop_if_not(attr(globals_ii, "resolved"))
-  
-      ## Adjust option 'future.globals.maxSize' to account for the fact that more
-      ## than one element is processed per future.  The adjustment is done by
-      ## scaling up the limit by the number of elements in the chunk.  This is
-      ## a "good enough" approach.
-      ## (https://github.com/HenrikBengtsson/future.apply/issues/8).
-      if (length(chunk) > 1L) {
-        globals_ii["...future.globals.maxSize"] <- list(globals.maxSize)
-        options(future.globals.maxSize = length(chunk) * globals.maxSize.default)
-        if (debug) mdebugf(" - Adjusted option 'future.globals.maxSize': %.0f -> %d * %.0f = %.0f (bytes)", globals.maxSize.default, length(chunk), globals.maxSize.default, getOption("future.globals.maxSize"))
-        on.exit(options(future.globals.maxSize = globals.maxSize), add = TRUE)
-      }
-      
-      ## Using RNG seeds or not?
-      if (is.null(seeds)) {
-        if (debug) mdebug(" - seeds: <none>")
-      } else {
-        if (debug) mdebugf(" - seeds: [%d] <seeds>", length(chunk))
-        globals_ii[["...future.seeds_ii"]] <- seeds[chunk]
-      }
-  
-      fs[[ii]] <- future(
-        expr, substitute = FALSE,
-        envir = future.envir,
-        stdout = future.stdout,
-        conditions = future.conditions,
-        globals = globals_ii, packages = packages_ii,
-        seed = future.seed,
-        label = labels[ii]
-      )
-      
-      if (debug) {
-        mdebug("Created future:")
-        mprint(fs[[ii]])
-      }
-      
-      ## Not needed anymore
-      rm(list = c("chunk", "globals_ii"))
-  
-      if (debug) mdebugf("Chunk #%d of %d ... DONE", ii, nchunks)
-    } ## for (ii ...)
-    if (debug) mdebugf("Launching %d futures (chunks) ... DONE", nchunks)
-  
-    ## 4. Resolving futures
-    if (debug) mdebugf("Resolving %d futures (chunks) ...", nchunks)
-  
-    ## Check for RngFutureCondition:s when resolving futures?
-    if (isFALSE(future.seed)) {
-      withCallingHandlers({
-        values <- local({
-          oopts <- options(future.rng.onMisuse.keepFuture = FALSE)
-          on.exit(options(oopts))
-          value(fs)
-        })
-      }, RngFutureCondition = function(cond) {
-        ## One of "our" futures?
-        idx <- NULL
-        
-        ## Compare future UUIDs or whole futures?
-        uuid <- attr(cond, "uuid")
-        if (!is.null(uuid)) {
-          ## (a) Future UUIDs are available
-          for (kk in seq_along(fs)) {
-            if (identical(fs[[kk]]$uuid, uuid)) idx <- kk
-          }
-        } else {        
-          ## (b) Future UUIDs are not available, use Future object?
-          f <- attr(cond, "future")
-          if (is.null(f)) return()
-          ## Nothing to do?
-          if (!isFALSE(f$seed)) return()  ## shouldn't really happen
-          for (kk in seq_along(fs)) {
-            if (identical(fs[[kk]], f)) idx <- kk
-          }
-        }
-        
-        ## Nothing more to do, i.e. not one of our futures?
-        if (is.null(idx)) return()
-        
-        ## Adjust message to give instructions relevant to this package
-        f <- fs[[idx]]
-        label <- f$label
-        if (is.null(label)) label <- "<none>"
-        message <- sprintf("UNRELIABLE VALUE: One of the %s iterations (%s) unexpectedly generated random numbers without declaring so. There is a risk that those random numbers are not statistically sound and the overall results might be invalid. To fix this, specify 'future.seed=TRUE'. This ensures that proper, parallel-safe random numbers are produced via the L'Ecuyer-CMRG method. To disable this check, use 'future.seed = NULL', or set option 'future.rng.onMisuse' to \"ignore\".", sQuote(.packageName), sQuote(label))
-        cond$message <- message
-        if (inherits(cond, "warning")) {
-          warning(cond)
-          invokeRestart("muffleWarning")
-        } else if (inherits(cond, "error")) {
-          stop(cond)
-        }
-      }) ## withCallingHandlers()
-    } else {
-      values <- value(fs)
-    }
+        ## Not needed anymore
+        rm(list = c("chunk", "globals_ii"))
     
-  
+        if (debug) mdebugf("Chunk #%d of %d ... DONE", ii, nchunks)
+      } ## for (ii ...)
+      if (debug) mdebugf("Launching %d futures (chunks) ... DONE", nchunks)
+    
+      ## 4. Resolving futures
+      if (debug) mdebugf("Resolving %d futures (chunks) ...", nchunks)
+    
+      ## Check for RngFutureCondition:s when resolving futures?
+      if (isFALSE(future.seed)) {
+        withCallingHandlers({
+          values <- local({
+            oopts <- options(future.rng.onMisuse.keepFuture = FALSE)
+            on.exit(options(oopts))
+            value(fs)
+          })
+        }, RngFutureCondition = function(cond) {
+          ## One of "our" futures?
+          idx <- NULL
+          
+          ## Compare future UUIDs or whole futures?
+          uuid <- attr(cond, "uuid")
+          if (!is.null(uuid)) {
+            ## (a) Future UUIDs are available
+            for (kk in seq_along(fs)) {
+              if (identical(fs[[kk]]$uuid, uuid)) idx <- kk
+            }
+          } else {        
+            ## (b) Future UUIDs are not available, use Future object?
+            f <- attr(cond, "future")
+            if (is.null(f)) return()
+            ## Nothing to do?
+            if (!isFALSE(f$seed)) return()  ## shouldn't really happen
+            for (kk in seq_along(fs)) {
+              if (identical(fs[[kk]], f)) idx <- kk
+            }
+          }
+          
+          ## Nothing more to do, i.e. not one of our futures?
+          if (is.null(idx)) return()
+          
+          ## Adjust message to give instructions relevant to this package
+          f <- fs[[idx]]
+          label <- sQuoteLabel(f)
+          message <- sprintf("UNRELIABLE VALUE: One of the %s iterations (%s) unexpectedly generated random numbers without declaring so. There is a risk that those random numbers are not statistically sound and the overall results might be invalid. To fix this, specify 'future.seed=TRUE'. This ensures that proper, parallel-safe random numbers are produced via a parallel RNG method. To disable this check, use 'future.seed = NULL', or set option 'future.rng.onMisuse' to \"ignore\".", sQuote(.packageName), label)
+          cond$message <- message
+          if (inherits(cond, "warning")) {
+            warning(cond)
+            invokeRestart("muffleWarning")
+          } else if (inherits(cond, "error")) {
+            stop(cond)
+          }
+        }) ## withCallingHandlers()
+      } else {
+        ## value() exits early if it detects a future with an error.
+        ## In future (>= 1.40.0), non-resolved futures will be automatically
+        ## canceled if there's an error.
+        value(fs)
+      }
+    }, interrupt = function(int) {
+      onInterrupt(int, fcn_name = fcn_name, debug = debug)
+    }, error = function(e) {
+      onError(e, futures = fs, debug = debug)
+    }) ## tryCatch()
+
     ## Not needed anymore
     rm(list = "fs")
   
@@ -319,7 +346,7 @@ future_xapply <- local({
     }
   
     if (debug) mdebugf("Reducing values from %d chunks ... DONE", nchunks)
-  
+
     values
   } ## future_xapply()
 })
